#!/usr/bin/env python3

import statistics
import json
import markdown_table
import argparse
import os
import parse

import numpy as np
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt

from math import sqrt

class MeasureSetHandler:

    report_description = ("This report was generated with the measure-analysis.py script.\nJSON formated measures "
                         "were imported in the measures folder using the import functionnality of the script, this report was then "
                         "generated using these measures.\nMetadata is included with the measures, such as the kernel "
                         "version used, the boot parameters passed, various others parameters specific to the measure, etc... "
                         "\nMeasures measuring the same propriety are grouped together in tables and graphs, and are identified "
                         "by their diverging metadatas. This is useful to analyse the effect of specific parameters on the  "
                         "measured propriety.\n\n "
                         "Currently, most of the measures need to be redone, as the code for sending the packets has changed "
                         "and gave different results.")
                         

    measures_dir = "measures"
    graphs_dir = "{}/graphs".format(measures_dir)
    measure_sets_file_name = "measure_sets.json"
    measure_sets_path = measures_dir + "/" + measure_sets_file_name
    ignored_metadata = ["board"]

    def __init__(self):
        with open(MeasureSetHandler.measure_sets_path) as measure_sets_file:
            self.measure_sets = json.load(measure_sets_file)

    def __str__(self):
        s = "List of measures:\n\n"
        id_strs = []
        for measure_set in self.measure_sets:
            id_strs.append("      {} : {}".format(measure_set, self.measure_sets[measure_set]['ids']))

        max_id_str_len = max([ len(s) for s in id_strs ])

        for i, measure_set in enumerate(self.measure_sets):

            next_id_str = self.measure_sets[measure_set]['next_id']
            next_id_str = "{} (Next id: {})\n".format(" " * (max_id_str_len + 5 - len(id_strs[i])), next_id_str)
            s+= id_strs[i] + next_id_str
        return s

    def save(self):
        
        remove_list = []
        for mtype in self.measure_sets:
            if len(self.measure_sets[mtype]['ids']) == 0:
                remove_list.append(mtype)
        for mtype in remove_list:
            del self.measure_sets[mtype]

        with open(MeasureSetHandler.measure_sets_path, 'w') as measure_sets_file:
            json.dump(self.measure_sets, measure_sets_file)

    def import_from_json(self, path, flat=False, cyclictest=False, user_input=False, file_input=None):

        measure_sets = []

        with open(path) as infile:

            if cyclictest:
                data = MeasureSet.parse_cyclictest(infile)
            else:
                data = json.load(infile)

            if file_input != None:
                with open(file_input) as metadata_file:
                    metadata = json.load(metadata_file)

            for measure_set_data in data['measure_sets']:

                measure_set = MeasureSet()
                measure_set.import_from_json_data(measure_set_data, flat=flat, user_input=user_input)

                if file_input != None:
                    measure_set.update_metadata(metadata, []);

                measure_sets.append(measure_set)

        return measure_sets

    def get_measure_set(self, measure_name):

        measure_path = "{}/{}.json".format(self.measures_dir, measure_name)
        measure_set = self.import_from_json(measure_path, False)[0]
        return measure_set

    def add_measure_set(self, measure_set):

        measure_set.compute_duration()
        mtype = measure_set.measure_type

        if mtype not in self.measure_sets:
            self.measure_sets[mtype] = {'ids': [], 'next_id': 0}

        next_id = self.measure_sets[mtype]['next_id']
        measure_file_name = "{}/{}{}.json".format(MeasureSetHandler.measures_dir, mtype, next_id)
        measure_set.export_to_json(measure_file_name)

        self.measure_sets[mtype]['ids'].append(next_id)
        self.measure_sets[mtype]['next_id'] += 1

        self.save()

        print("Saved measure as {}{}".format(mtype, next_id))

    def remove_measure_set(self, mtype, mid):
        if mtype in self.measure_sets and mid in self.measure_sets[mtype]['ids']:
            self.measure_sets[mtype]['ids'].remove(mid)
            measure_file_name = "{}/{}{}.json".format(MeasureSetHandler.measures_dir, mtype, mid)
            os.remove(measure_file_name)
            self.save()
            print("Removed measure {}{}".format(mtype, mid))
        else:
            print("{}{} doesn't exist".format(mtype, mid))

    def remove_mtype(self, mtype):
        measures_removed = 0

        if mtype not in self.measure_sets:
            print("No type named {}".format(mtype))
            return

        ids = [index for index in self.measure_sets[mtype]['ids']]
        for mid in ids:
            self.remove_measure_set(mtype, mid)
            measures_removed += 1

        if measures_removed:
            print("Removed {} measures of the type {}".format(measures_removed, mtype))
        else:
            print("No measures were removed")

    def update_metadata(self, mtype, mid, metadata, removed_metadata):
        if mtype in self.measure_sets and mid in self.measure_sets[mtype]['ids']:
            measure_set = self.get_measure_set("{}{}".format(mtype, mid))
            measure_set.update_metadata(metadata, removed_metadata)
            measure_file_name = "{}/{}{}.json".format(MeasureSetHandler.measures_dir, mtype, mid)
            measure_set.export_to_json(measure_file_name)
        else:
            print("{}{} doesn't exist".format(mtype, mid))

    def modify_metadata(self, mtype, mid=-1):
        update_all = (mid == -1)
        if mtype in self.measure_sets:
            measure_ids = self.measure_sets[mtype]['ids']
        else:
            print("There are no measures of type {}".format(mtype))
            return
        if update_all:
            mid = measure_ids[0]

        if mtype in self.measure_sets and mid in measure_ids:
            measure_set = self.get_measure_set("{}{}".format(mtype, mid))
            metadata, removed_metadata = measure_set.input_metadata()
            self.update_metadata(mtype, mid, metadata, removed_metadata)
            if update_all:
                for measure_id in measure_ids:
                    self.update_metadata(mtype, measure_id, metadata, removed_metadata)
        else:
            print("{}{} doesn't exist".format(mtype, mid))

    def remove_all(self):
        for mtype in self.measure_sets:
            while True:
                if len(self.measure_sets[mtype]['ids']) == 0:
                    break
                mid = self.measure_sets[mtype]['ids'][0]
                print("    Deleting {}{}...".format(mtype, mid))
                self.remove_measure_set(mtype, mid)
        print("Removed all measures")

    def generate_graphs(self, metadata_masks):

        # For each type of measure set
        for mtype in self.measure_sets:

            # List all the measures of this type
            measures = []
            mids = []
            for mid in self.measure_sets[mtype]['ids']:
                measure = self.get_measure_set("{}{}".format(mtype, mid))
                # Measures with lost packets are not relevant to graph
                if 'lost_packets' not in measure.metadata or measure.metadata['lost_packets'] == 0:
                    measures.append(measure)
                    mids.append(mid)
            if len(measures) == 0:
                continue

            # Get the type and number of properties, which are all identical for
            # measures of the same type
            props_type = measures[0].props_type
            nb_props = len(measures[0].props)

            # For each property draw a graph
            for i in range(nb_props):
                prop_name = measures[0].props_names[i]

                subplots = []
                metadata_infos = []

                # Superimpose all the different measures of the same type and property type
                for j, measure in enumerate(measures):
                    mid = mids[j]
                    graph_name = "{}{}".format(mtype, mid)

                    # Generate the plot name from the metadata mask
                    metadata_info = ["{}".format(measure.metadata[metadata_name]) for metadata_name in metadata_masks[mtype]]
                    metadata_infos.append(", ".join(metadata_info))

                    if props_type == 'histogram':
                        subplots.append(measure.generate_histogram(i, metadata_info))
                    else:
                        subplots.append(measure.generate_chrono_graph(i, metadata_info))

                fig, ax = plt.gcf(), plt.gca()

                # Set meta graph information 
                if props_type == 'histogram':
                    ax.set_xlabel('Latency (us)')
                    ax.set_ylabel('Percentage of latency samples (%)')
                    ax.set_title('{}, {} histogram'.format(mtype, prop_name))
                else:
                    ax.set_xlabel('Time (us)')
                    ax.set_ylabel(prop_name)
                    ax.set_title('{}, {} graph'.format(mtype, prop_name))

                plt.legend(subplots, metadata_infos)
                #ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)

                fig.set_size_inches(15.0, 8.0)


                # Save the graph
                plt.savefig("{}/{}{}.png".format(MeasureSetHandler.graphs_dir, mtype, prop_name))
                plt.clf()

    def generate_metadata_text(self):

        metadata_masks = {}
        common_metadata_strs = {}

        for mtype in self.measure_sets:
            metadata_mask = []
            measures = []
            for mid in self.measure_sets[mtype]['ids']:
                measures.append(self.get_measure_set("{}{}".format(mtype, mid)))

            if len(measures) == 0:
                continue
            first_metadata = measures[0].metadata

            # Generate the metadata mask, by grouping the identical metadata
            for measure in measures[1:]:
                for metadata_name in measure.metadata:
                    if metadata_name not in self.ignored_metadata and metadata_name != 'lost_packets':
                        # If it is not already in the metadata mask
                        if metadata_name not in metadata_mask:
                            # If there are two different metadata, they are added to the mask
                            if measure.metadata[metadata_name] != first_metadata[metadata_name]:
                                metadata_mask.append(metadata_name)

            metadata_masks[mtype] = metadata_mask

            # Compute the identical metadata
            common_metadata = []
            for metadata_name in first_metadata:
                if metadata_name not in self.ignored_metadata and metadata_name != 'lost_packets':
                    if metadata_name not in metadata_mask:
                        if metadata_name in MeasureSet.abbreviations:
                            common_metadata.append("{}: {}".format(MeasureSet.abbreviations[metadata_name], first_metadata[metadata_name]))
                        else:
                            common_metadata.append("{}: {}".format(metadata_name, first_metadata[metadata_name]))
            common_metadata_strs[mtype] = ", ".join(common_metadata)

        return (metadata_masks, common_metadata_strs)

    def generate_report(self, include_graphs=True):

        metadata_masks, common_metadata_strs = self.generate_metadata_text()

        if include_graphs:
            self.generate_graphs(metadata_masks)

        with open("measure-report.md", 'w+') as report:

            report.write("## Measurements\n\n")
            report.write("{}\n\n".format(MeasureSetHandler.report_description))

            report.write("### Description of the metadata associated to each test\n\n")
            for abbr_name in MeasureSet.metadata_descriptions:
                if abbr_name not in self.ignored_metadata:
                    report.write("* **{}**: {}\n".format(abbr_name, MeasureSet.metadata_descriptions[abbr_name]))
            report.write("\n")

            for mtype in self.measure_sets:
                need_header = True
                props_lens = []

                report.write("### {} results\n\n".format(mtype))

                # mtype description
                if mtype in MeasureSet.descriptions:
                    report.write('{}\n\n'.format(MeasureSet.descriptions[mtype]))

                # List all the measures of this type
                measures = []
                for mid in self.measure_sets[mtype]['ids']:
                    measure = self.get_measure_set("{}{}".format(mtype, mid))
                    measures.append(measure)

                if len(measures) == 0:
                    continue

                packet_measures = 'lost_packets' in measures[0].metadata

                if packet_measures:
                    measures.sort(key=lambda x:x.metadata['lost_packets'])

                first_metadata = measures[0].metadata

                # Write the identical metadata before the table
                report.write("**Common test metadata:** {}\n\n".format(common_metadata_strs[mtype]))

                for measure in measures:

                    if need_header:

                        table_str, props_lens = measure.generate_table(packet_measures, headers=True, metadata_mask=metadata_masks[mtype])
                        report.write(table_str)
                        need_header = False
                    else:
                        report.write(measure.generate_table(packet_measures, headers=False, props_lens=props_lens, metadata_mask=metadata_masks[mtype])[0])

                    report.write("\n")

                # Include the graphs
                if include_graphs:
                    for i in range(len(measures[0].props)):
                        report.write('\n![alt text]({}/graphs/{}{}.png "{} Graph")\n'.format(MeasureSetHandler.measures_dir,
                                                                                             mtype,
                                                                                             measures[0].props_names[i],
                                                                                             mtype,
                                                                                             measures[0].props_names[i]))
                report.write("\n")
            report.write("\n")

class MeasureSet:

    abbreviations = {
                      'ker': 'Linux kernel version',
                      'prio': 'Task priority',
                      'i': 'Interval',
                      'board': 'Board name',
                      'boot_p': 'Boot Parameters',
                      'delta': 'ETF qdisc delta',
                      'load': 'Device and processor load',
                      'duration': 'Test duration',
                      'speed': 'Ethernet speed in Mb/s',
                    }

    metadata_descriptions = {
                              'ker': 'Linux kernel version used',
                              'prio': 'Priority of the real-time thread',
                              'i': 'Interval between each packet transmission',
                              'board': 'Name of the board used, one of three : Emerald, Slate or Onyx. (All boards are identical)',
                              'boot_p': 'Boot Parameters passed to u-boot',
                              'delta': 'ETF qdisc delta value',
                              'load': 'External device and processor load during the test',
                              'duration': 'Duration of the test',
                              'etf_offset': 'When a packet\'s txtime is scheduled, the txtime set is the current time offset by etf_offset',
                              'route': "How the boards are connected. 'switch' means they are connected by a switch (cheap, unmanaged), 'E2E'"
                                       "means they are directly connected end to end",
                            }

    descriptions = {
                      'cyclictest_wake-up_latency': "Uses cyclictest from the rt-tests test suite to measure wake-up latency. "
                                                     "A real-time thread is run on CPU1, and wakes up at a regular interval "
                                                     "(specified by the interval parameter) using clock_nanosleep. It then uses "
                                                     "clock_gettime and computes the difference between the scheduled wake-up "
                                                     "time and the effective wake-up time measured by clock_gettime. "
                                                     "The command used is: cyclictest -p `prio` -a1 -t1 -n -h 200 -q -i `interval`",

                      'packet_jitter':              "An UDP packet is periodically sent from one board to another using a real time "
                                                     "thread. The receiving board calculates the intervals between the packets "
                                                     "it receives, and sees how much it differs from the scheduled interval.",

                      'packet_recv_timestamps':     "An UDP packet is periodically sent from one board to another using a real time "
                                                     "thread. The receiving board uses the SO_TIMESTAMPING option to see when "
                                                     "the packet entered the kernel, and generates timestamps with clock_gettime "
                                                     "once the packets enters userspace to compute the time the packet spent in "
                                                     "kernel space.",

                      'packet_send_timestamps':      "Similar to packet_recv_timestamps, but on the transmitting board. ",

                      'packet_rtt':                  "An UDP packet is periodically sent from one board to another and sent back as "
                                                      "soon as it arrives using a real time thread. The round trip time is computed "
                                                      "with clock_gettime on the transmitting board.",
                  }

    def __init__(self):

        # Default values
        self.metadata = {}
        self.metadata['board'] = "Emerald"
        self.metadata['ker'] = "4.19"
        self.metadata['boot_p'] = "isolcpus"
        self.metadata['i'] = "200us"
        self.metadata['delta'] = "none"
        self.metadata['prio'] = "99"
        self.metadata['load'] = "ssh"
        self.metadata['duration'] = "24h"
        self.metadata['qdisc'] = "pfifo_fast"
        self.metadata['etf_offset'] = "none"
        self.metadata['route'] = "switch"
        self.metadata['speed'] = "100"

        self.props = []
        self.props_name = []
        self.units = []

    def __str__(self):

        return "Cols: " + str(self.props) + "\n"

    def merge_metadata(metadata, new_metadata, removed_metadata):
        metadata.update(new_metadata)
        for metadata_name in removed_metadata:
            metadata.pop(metadata_name, None)
        return metadata

    def update_metadata(self, metadata, removed_metadata):
        self.metadata = MeasureSet.merge_metadata(self.metadata, metadata, removed_metadata)

    def input_metadata(self):

        metadata = {}
        removed_metadata = []

        metadata_name = ""
        while True:

            merged_metadata = MeasureSet.merge_metadata(self.metadata, metadata, removed_metadata)
            print("Current metadata:\n")
            for metadata_name in merged_metadata:
                print("   {}: {}".format(metadata_name, merged_metadata[metadata_name]))

            metadata_name = input('Enter metadata name (type "done" to exit): ')
            if metadata_name == "done":
                break
            metadata_value = input('Enter metadata value (type "done" to exit, "delete" to delete current metadata, "cancel" to cancel current metadata): ')
            if metadata_value == "done":
                break
            if metadata_value == "cancel":
                continue
            if metadata_value == "delete":
                removed_metadata.append(metadata_name)
                continue
            metadata[metadata_name] = metadata_value

        return (metadata, removed_metadata)

    def add_metadata(self, measure_type, units, middle, metadata, user_input=False):

        self.measure_type = measure_type
        self.units = units
        self.middle = int(middle)
        self.metadata.update(metadata)

        if user_input:
            new_metadata, removed_metadata = self.input_metadata()
            self.update_metadata(new_metadata, removed_metadata)

    def add_chronological(self, props_names, props):

        self.props = props
        self.props_names = props_names
        self.props_type = 'chronological'

        self.max = [max(prop) for prop in props]
        self.min = [min(prop) for prop in props]
        self.avg = [statistics.mean(prop) for prop in props]
        self.dev = [sqrt(statistics.variance(prop)) for prop in props]

    def add_histogram(self, props_names, props):

        self.props = props
        self.props_names = props_names
        self.props_type = 'histogram'

        self.max = []
        self.min = []
        self.avg = []
        self.dev = []

        self.nb_measures = []

        for prop in props:

            enumerate_prop = list(enumerate(prop))

            self.max.append(max(list(map(lambda x: (x[0] - self.middle) if x[1] else 0, enumerate_prop))))
            self.min.append(min(map(lambda x: (x[0] - self.middle) if x[1] else self.max[-1], enumerate_prop)))

            sum_prop = sum(prop)
            self.nb_measures.append(sum_prop)

            avg = sum(map(lambda x: abs(x[0] - self.middle)*x[1], enumerate_prop)) / sum_prop
            avg_centered = sum(map(lambda x: x[0]*x[1], enumerate_prop)) / sum_prop
            var = 0
            for x, p in enumerate_prop:
                var += p * x**2
            var /= sum_prop
            var -= avg_centered**2

            self.avg.append(avg)
            self.dev.append(sqrt(var))
            
    def export_to_json(self, path):

        with open(path, 'w') as outfile:
            json.dump({'measure_sets': [
                         {'measure_type': self.measure_type,
                          'props_names': self.props_names,
                          'units': self.units,
                          'middle': self.middle,
                          'props': self.props,
                          'props_type': self.props_type,
                          'metadata': self.metadata}
                         ]}, outfile)

    def parse_cyclictest(infile):

        data = {}
        measure_set = {}
        measure_set['measure_type'] = 'cyclictest_wake-up_latency'
        measure_set['props_type'] = 'histogram'
        measure_set['props_names'] = ['wake-up latency']
        measure_set['units'] = ['us']
        measure_set['props'] = [[]]
        measure_set['metadata'] = {}

        lines = [line for line in infile]

        for line in lines[2:]:
            if line[0] == '#':
                break
            i, x = parse.parse('{:d} {:d}', line)
            measure_set['props'][0].append(x)

        data['measure_sets'] = [measure_set]

        return data

    def compute_duration(self):

        nb_packets = sum(self.props[0])
        nb_us = int(self.metadata['i'][:-2]) * nb_packets
        hours = int((nb_us / 10**6) / 3600)
        minutes = int((nb_us / 10**6) / 60) % 60
        self.metadata['duration'] = "{}h{}".format(hours, minutes)

    def import_from_json_data(self, data, flat=False, user_input=False):

        measure_type = data['measure_type']
        units = data['units']
        metadata = data['metadata']
        if 'middle' in data:
            middle = data['middle']
        else:
            middle = 0

        self.add_metadata(measure_type, units, middle, metadata, user_input)

        props_names = data['props_names']

        if data['props_type'] == 'histogram':

            props = data['props']
            self.add_histogram(props_names, props)

        else:
            if flat:

                values = data['values']

                nb_props = len(props_names)
                props = [[] for c in range(nb_props)]

                for i,value in enumerate(values):
                    props[i % nb_props].append(value)
            else:
                props = data['props']

            self.add_chronological(props_names, props)

    def generate_histogram(self, i, name):

        bins = [i for i in range(self.min[i], self.max[i] + 1)]
        vals = self.props[i][self.min[i] + self.middle : self.max[i] + self.middle + 1]
        vals = [ (x * 100) / sum(vals) for x in vals]

        min_val = max(vals) / 140

        vals = list(map(lambda x: min_val if x != 0 and x < min_val else x, vals))

        return plt.bar(bins, vals, alpha=0.4)

    def generate_chrono_graph(self, i, name):
        prop = self.props[i]
        x = [i for i in range(len(prop))]
        return plt.plot(x, prop)

    def generate_table(self, packet_measures, headers=True, values=True, metadata_mask=[], props_lens=[]):

        if headers == False and values == False:
            return ""

        table = []

        if headers:

            # Table headers
            headers = ["Minimum", "Maximum", "Average", "Standard deviation"]
            if packet_measures:
                headers.append("Lost packets")

            # Add a metadata information column if the metadata_mask isn't empty
            if metadata_mask != []:
                table += [["Metadata"] + headers]
                table += [["**" + ", ".join(metadata_mask) + "**"] + ["**" + " - ".join(self.props_names) + "**"] * len(headers)]
            else:
                table +=  [headers]
                table += [["**" + " - ".join(self.props_names) + "**"] * len(headers)]

            if packet_measures:
                table[-1][-1] = ""

        if values:
            m = [self.min, self.max, self.avg, self.dev]

            values = [[ (format(m[i][j], '.4f') + self.units[j]) for j in range(len(m[0]))] for i in range(len(m))] 

            if metadata_mask != []:
                metadata_info = ["{}".format(self.metadata[metadata_name]) for metadata_name in metadata_mask]
                table += [["{} ({})".format(", ".join(metadata_info), ", ".join(map(str, self.nb_measures)))] +
                          [" - ".join(values[i]) for i in range(len(values))]]
            else:
                table += [[" - ".join(values[i]) for i in range(len(values))]]
            if packet_measures:
                table[-1] += [self.metadata['lost_packets']]

        if props_lens == []:
            props_lens = [max([len(table[i][j]) for i in range(len(table))]) for j in range(len(table[0]))]
        table = [[ table[i][j].ljust(props_lens[j]) for j in range(len(table[0]))] for i in range(len(table))]

        table_str = ""

        if headers:
            table_str += " | ".join(table[0]) + "\n"
            table_str += " | ".join([ ("-" * props_lens[i]) for i in range(len(props_lens)) ]) + "\n"
            if values:
                table_str += "\n".join([" | ".join(line) for line in table[1:]])
        else:
            table_str += "\n".join([" | ".join(line) for line in table])

        return (table_str, props_lens)

def name_to_mtype_mid(file_name):
        mid_start = 0
        for c in range(len(file_name)-1, -1, -1):
            if not str.isdigit(file_name[c]):
                mid_start = c+1
                break

        mtype = file_name[:mid_start]
        mid = int(file_name[mid_start:])

        return (mtype, mid)


def parse_args():
    parser = argparse
    parser = argparse.ArgumentParser(description='Measure analysis')
    parser.add_argument('-i', nargs=1, required=False, help='import file')
    parser.add_argument('-c', action='store_true', required=False, help='parse cyclictest histogram')
    parser.add_argument('-I', nargs=1, required=False, help='import metadata file')
    parser.add_argument('--remove', nargs=1, required=False, help='remove a measure')
    parser.add_argument('--remove-type', nargs=1, required=False, help='remove all measures from a type')
    parser.add_argument('--remove-all', action='store_true', help='remove all measure sets')
    parser.add_argument('-m', nargs=1, required=False, help='modify the metadata of a measure')
    parser.add_argument('-M', nargs=1, required=False, help='modify the metadata of all measures of a type')
    parser.add_argument('-t', nargs='?', const='input_file', required=False, help='generate table')
    parser.add_argument('-R', action='store_true', required=False, help='generate full measure report')
    parser.add_argument('-G', action='store_true', required=False, help='generate all graphs')
    parser.add_argument('-s', action='store_true', help='show measures')

    args = parser.parse_args()

    ms_handler = MeasureSetHandler()

    if args.remove is not None:
        
        mtype, mid = name_to_mtype_mid(args.remove[0])
        ms_handler.remove_measure_set(mtype, mid)

    if args.remove_type is not None:
        
        ms_handler.remove_mtype(args.remove_type[0])

    if args.m is not None:

        mtype, mid = name_to_mtype_mid(args.m[0])
        ms_handler.modify_metadata(mtype, mid)
    
    if args.M is not None:

        mtype = str(args.M[0])
        ms_handler.modify_metadata(mtype)
    
    if args.i is not None:

        if args.I is not None:
            file_input = args.I[0]
            user_input = False
        else:
            file_input = None
            user_input = True

        if args.c:
            measure_sets = ms_handler.import_from_json(args.i[0], cyclictest=True, user_input=user_input, file_input=file_input)
        else:
            measure_sets = ms_handler.import_from_json(args.i[0], flat=True, user_input=user_input, file_input=file_input)

        for measure_set in measure_sets:
            ms_handler.add_measure_set(measure_set)

    elif args.t is not None and args.t != "input_file":
        measure_set = ms_handler.get_measure_set(args.t)
        print(measure_set.generate_table(True, True, {'board', 'linux_version', 'boot_params'})[0])

    if args.R:
        ms_handler.generate_report()

    if args.G:
        ms_handler.generate_graphs()

    if args.remove_all:
        confirm = input("Are you sure all measure sets should be removed ? [Yes] / [No]: ")
        if confirm == "Yes":
            ms_handler.remove_all()

    if args.s:
       print(ms_handler)

parse_args()
