/*
 * Real time packet receiving server
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <inttypes.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <unistd.h>

#include "common.h"
#include "gpio.h"
#include "recv_packet.h"
#include "send_packet.h"
#include "tracer.h"

// Structs

enum TSNTask { RECV_PACKET_TASK, RTT_TASK, XDP_TASK };

typedef struct thread_param {
  int interval;
  int priority;
  int affinity_cpu;
  uint64_t latency_threshold;

  int emit_signal;

} thread_param_t;

typedef struct main_params {
  int refresh_rate;
  int verbose;
  int enable_tracing;
  int enable_xdp_tracing;
  int enable_graph;
} main_param_t;

static void process_options(int argc, char *argv[]);
static void print_histograms();
static void sighand(int sig_num);

// Static variables

static uint64_t kernel_latency_hist[MAX_KERNEL_LATENCY];
static uint64_t jitter_hist[MAX_JITTER];
static main_param_t main_params;
static thread_param_t thread_params;
static ingress_param_t ingress_params;
static egress_param_t egress_params;
static egress_stat_t egress_stats = {.min_kernel_latency = INT_MAX};
static ingress_stat_t ingress_stats = {.min_kernel_latency = INT_MAX,
                                       .min_interval = INT_MAX};
static int enable_histograms;
static int enable_timestamps;
static enum TSNTask tsn_task;
static struct timespec measures_start;
static struct timespec measures_end;
static struct timespec emit_signal_next;
static pthread_mutex_t emit_signal_mutex;
static pthread_cond_t emit_signal_ts_received;

static void help(char *argv[]) {
  printf(
      "Usage: %s [-f IF -a CPU -p PRIO -i USEC -r USEC] [-b CLIENT_IP] [-d "
      "BUF_LEN -cgtv] [-T LATENCY_THRESHOLD -G] [-x POLL_MODE -X]\n\n"
      "       -h                 Show help\n"
      "       -f IF              Set the network interface to be used\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -i USEC            RT thread wake-up interval (default: 10ms)\n"
      "       -r USEC            non-RT main thread refresh interval\n"
      "       -d BUF_LEN         Set the length of tx buffer\n"
      "       -c                 Receive timestamp and emit signal\n"
      "       -b CLIENT_IP       Server side RTT\n"
      "       -g                 Print histograms to sdtout on exit\n"
      "       -t                 Enable timestamps\n"
      "       -x POLL_MODE       Use AF_XDP sockets\n"
      "          POLL_MODE: 0 for polling, 1 for combination of both\n"
      "       -X                 Trace during XDP packet reception\n"
      "       -T THRESHOLD       Enable tracing until THRESHOLD is reached\n"
      "       -G                 Enable function_graph tracer, used with -T\n"
      "       -v                 Verbose\n"
      "\n",
      argv[0]);
}

static void *emit_signal_thread(void *p) {
  (void)p;
  cpu_set_t mask;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  pthread_mutex_lock(&emit_signal_mutex);

  for (;;) {
    pthread_cond_wait(&emit_signal_ts_received, &emit_signal_mutex);
    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &emit_signal_next, NULL);

    toggle_gpio(49);
    printf("Toggled GPIO\n");
  }

  pthread_mutex_unlock(&emit_signal_mutex);
  pthread_exit(NULL);
}

/* Real-time thread:
 *   - Measures intervals between packet receptions
 */
static void *tsn_thread(void *p) {
  (void)p;
  struct timespec current, previous;
  cpu_set_t mask;
  char tracemark_message[128];

  // Set thread CPU affinity
  if (thread_params.affinity_cpu) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_MONOTONIC, &measures_start);

  if (tsn_task == XDP_TASK) setup_poll_fd();

  // Packet receiving loop
  for (ingress_stats.packets_received = 0;; ingress_stats.packets_received++) {
    // RTT
    if (tsn_task == RTT_TASK) {
      recv_udp_packet();
      send_udp_packet("", 0);

      // Receive packet
    } else if (tsn_task == RECV_PACKET_TASK || tsn_task == XDP_TASK) {
      // Receive UDP or XDP packet
      if (tsn_task == RECV_PACKET_TASK)
        recv_udp_packet();
      else
        recv_xdp_packet();

      clock_gettime(CLOCK_MONOTONIC, &current);
      recv_xdp_cleanup();

      // Emit signal
      if (thread_params.emit_signal) {
        uint64_t emit_signal_t = *((uint64_t *)ingress_stats.data);

        pthread_mutex_lock(&emit_signal_mutex);
        emit_signal_next = uint_to_ts(emit_signal_t);
        pthread_cond_signal(&emit_signal_ts_received);
        pthread_mutex_unlock(&emit_signal_mutex);
      }

      // Update stats
      if (ingress_stats.packets_received) {
        int interval_us = calcdiff_ns(current, previous) / 1000;

        ingress_stats.min_interval =
            _min_(interval_us, ingress_stats.min_interval);
        ingress_stats.max_interval =
            _max_(interval_us, ingress_stats.max_interval);
        ingress_stats.avg_interval =
            (ingress_stats.avg_interval * ingress_stats.packets_received +
             interval_us) /
            (ingress_stats.packets_received + 1);

        // Histogram
        if (enable_histograms) {
          int dist_to_interval = interval_us - (thread_params.interval / 1000);
          dist_to_interval += MAX_JITTER / 2;

          if (dist_to_interval > ((int)MAX_JITTER) || dist_to_interval < 0)
            ingress_stats.high_jitter++;
          else
            jitter_hist[dist_to_interval]++;
        }
      }

      // If the latency hits the tracing threshold, stop tracing
      if (main_params.enable_tracing &&
          (ingress_stats.max_interval >
           ((int)thread_params.latency_threshold))) {
        sprintf(tracemark_message, "Jitter threshold hit: %dus\n",
                ingress_stats.max_interval);
        tracemark(tracemark_message);
        tracing(0);
        break;
      }

      previous = current;
    }
  }

  return NULL;
}

static void create_thread(void *(*thread_function)(void *)) {
  pthread_t thread;
  pthread_attr_t attr;
  struct sched_param param;

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  // Use scheduling parameters of attr
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, thread_function, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create packet receiving thread");
}

/* Main thread:
 *   - Has non-real time priority
 *   - Handles the IO and creates real time threads
 */
int main(int argc, char *argv[]) {
  // Initial values
  ingress_stats.min_interval = INT_MAX;
  ingress_stats.avg_interval = 0;
  ingress_stats.max_interval = 0;
  ingress_stats.min_kernel_latency = INT_MAX;
  ingress_stats.avg_kernel_latency = 0;
  ingress_stats.max_kernel_latency = 0;
  ingress_stats.packets_received = 0;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.priority = 99;
  thread_params.emit_signal = 0;
  thread_params.affinity_cpu = 0;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  main_params.enable_tracing = 0;
  main_params.enable_graph = 0;
  enable_timestamps = 0;
  enable_histograms = 0;
  tsn_task = RECV_PACKET_TASK;
  ingress_params.tx_buffer_len = 1024;

  // Process bash options
  process_options(argc, argv);

  ingress_params.use_timestamps = enable_timestamps;
  ingress_params.interval = thread_params.interval;

  // Init histograms
  if (enable_histograms) {
    memset(kernel_latency_hist, 0, sizeof(kernel_latency_hist));
    memset(jitter_hist, 0, sizeof(jitter_hist));
  }

  // Enable ftrace
  if (main_params.enable_tracing || main_params.enable_xdp_tracing) open_fds();

  // Catch breaks with sighand to print the histograms or exit XDP
  init_signals(sighand);

  // Initialize the XDP or UDP packet receiving socket
  if (tsn_task == XDP_TASK)
    init_xdp_recv(&ingress_params);
  else
    init_udp_recv(&ingress_params, &ingress_stats, enable_histograms,
                  kernel_latency_hist);

  // Initialize the UDP packet sending socket if RTT is measured
  if (tsn_task == RTT_TASK)
    init_udp_send(&egress_params, &egress_stats, 0, 0, NULL);

  if (thread_params.emit_signal) {
    pthread_mutex_init(&emit_signal_mutex, NULL);
    pthread_cond_init(&emit_signal_ts_received, NULL);

    enable_gpio(49);
  }

  create_thread(tsn_thread);
  create_thread(emit_signal_thread);

  // Verbose loop
  for (;;) {
    usleep(main_params.refresh_rate);

    // Stats printing
    if (main_params.verbose && ingress_stats.packets_received > 1 &&
        (tsn_task == RECV_PACKET_TASK || tsn_task == XDP_TASK)) {
      int jitter = ingress_stats.max_interval - ingress_stats.min_interval;

      printf("%9" PRIu64 ": J: %5d, I (10us): %3d %3d %3d [%3d]",
             ingress_stats.packets_received, jitter,
             ingress_stats.min_interval / 10, ingress_stats.avg_interval / 10,
             ingress_stats.max_interval / 10, (int)ingress_stats.high_jitter);

      if (enable_timestamps)
        printf(", K: %4d %4d %4d [%3d]\n", ingress_stats.min_kernel_latency,
               ingress_stats.avg_kernel_latency,
               ingress_stats.max_kernel_latency,
               (int)ingress_stats.high_kernel_latency);
      else
        printf("\n");

      printf("\033[%dA", 1);
    }

    if (main_params.enable_tracing &&
        (ingress_stats.max_interval >= ((int)thread_params.latency_threshold)))
      break;
  }

  pthread_mutex_destroy(&emit_signal_mutex);
  pthread_cond_destroy(&emit_signal_ts_received);
  close_xdp_socket();
}

/* Print histograms in .json format
 */
static void print_histograms() {
  uint64_t duration;
  int duration_hour, duration_minutes, interval;
  int max_latency, max_jitter, min_jitter;

  clock_gettime(CLOCK_MONOTONIC, &measures_end);

  duration = calcdiff_ns(measures_end, measures_start);
  duration_hour = duration / NSEC_PER_SEC / 3600;
  duration_minutes = duration / NSEC_PER_SEC / 60 - duration_hour * 60;

  interval = thread_params.interval / 1000;

  max_jitter = histogram_max(jitter_hist, MAX_JITTER - 1);
  min_jitter = histogram_min(jitter_hist, MAX_JITTER - 1);

  if (enable_timestamps) {
    max_latency = histogram_max(kernel_latency_hist, MAX_KERNEL_LATENCY - 1);

    printf(
        "Duration: %dh%d\n"
        "Lost packets: %" PRIu64 "\n",
        duration_hour, duration_minutes, ingress_stats.high_jitter);

    for (int j = 0; j < max_latency; j++)
      printf("%" PRIi64 "\n", kernel_latency_hist[j]);
  }

  printf(
      "Middle: %d\n"
      "Lost packets\": %" PRIu64 "\n",
      MAX_JITTER / 2 - min_jitter, duration_hour, duration_minutes,
      ingress_stats.high_jitter);

  for (int j = min_jitter; j < max_jitter; j++)
    printf("%" PRIi64 "\n", jitter_hist[j]);
}

static void sighand(int sig_num) {
  (void)sig_num;

  if (enable_histograms) {
    print_histograms();
  }

  if (tsn_task == XDP_TASK) {
    close_xdp_socket();
  }

  if (ingress_stats.high_jitter)
    fprintf(stderr, "%d packets were lost\n", ((int)ingress_stats.high_jitter));

  exit(EXIT_SUCCESS);
}

/* Process bash options
 */
static void process_options(int argc, char *argv[]) {
  int network_if_specified = 0;

  for (;;) {
    int c = getopt(argc, argv, "a:b:cd:f:ghi:p:r:tvx:XT:G");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 'b':
        tsn_task = RTT_TASK;
        strcpy(egress_params.server_ip, optarg);
        break;
      case 'c':
        thread_params.emit_signal = 1;
        break;
      case 'd':
        ingress_params.tx_buffer_len = atoi(optarg);
        if (ingress_params.tx_buffer_len < 1) {
          fprintf(stderr, "BUF_LEN should be greater than 1\n");
          exit(EXIT_FAILURE);
        }
        break;
      case 'f':
        network_if_specified = 1;
        strcpy(ingress_params.network_if, optarg);
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'i':
        thread_params.interval = atoi(optarg) * 1000;
        break;
      case 'g':
        enable_histograms = 1;
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 'r':
        main_params.refresh_rate = atoi(optarg);
        break;
      case 't':
        enable_timestamps = 1;
        break;
      case 'v':
        main_params.verbose = 1;
        break;
      case 'x':
        tsn_task = XDP_TASK;
        ingress_params.xdp_polling_mode = atoi(optarg);
        break;
      case 'X':
        main_params.enable_xdp_tracing = 1;
        break;
      case 'T':
        main_params.enable_tracing = 1;
        thread_params.latency_threshold = atoi(optarg);
        break;
      case 'G':
        main_params.enable_graph = 1;
        break;
    }
  }

  if (!network_if_specified) {
    fprintf(stderr, "You need to specifiy an network interface\n");
    help(argv);
    exit(EXIT_FAILURE);
  }

  if (argc != optind) {
    fprintf(stderr, "Too many arguments\n");
    help(argv);
    exit(EXIT_FAILURE);
  }
}
